; The smallest PIC controlling a simple 2x16 LCD

; Made by 'zuisti' 2007.07.02 (zuisti@yahoo.com)

;  This is my first work with a PIC system 
;       (and ... in Proteus only :-). 

;  My aim was to control a simple 2x16 char LCD and a traffic light,
;  and query two switches, all with the smallest PIC (12c508). 

;  The challenge was to fit the required code and circuitry to this 8 pin chip. 


;  This code uses the two wire LCD interface, copyright by 
;    Myke Predko (http://www.myke.com/lcd.htm). Thanks !!

;  Shift Data Order:
;  Bit 1 - Always High	(Gate for LCD "E")
;  Bit 2 - RS Bit	(0 for instr, 1 for char)
;  Bit 3 - LCD D7	(High nybble !)
;  Bit 4 - LCD D6
;  Bit 5 - LCD D5
;  Bit 6 - LCD D4

  LIST P=12C508, R=DEC          ;  12C508 (runs at 4 MHz), def: decimal

  errorlevel 0,-305
; Supressed Message[305]: Using default destination of 1 (f).

  INCLUDE "p12c508.inc"

;  Register File Usage (8 bit variables)

 CBLOCK 0x007			;  Start variables after end of the uC regs

cgoto_num			;  controls computed gotos (returns)
Offset				;  in place of FSR (this is a 5 bit reg).
TempChar			;  Used When Sending Out Data to the LCD
TempNybble			;  Temporary Value to "NybbleOut"
AnimSpeed			;  Animation speed (smaller is faster)
AnimCount			;  Animation loop counter
Count				
Count1
Delay1
Delay2

 ENDC

;  Defines

#DEFINE	Data	GPIO,0
#DEFINE	Clock	GPIO,1
#DEFINE	GreenL	GPIO,2
#DEFINE	RedL  	GPIO,4
#DEFINE	YellowL	GPIO,5

#DEFINE	SW1_ContrMode	GPIO,3	;  GP3 is input only

#DEFINE	GreenM	0x4
#DEFINE	RedM  	0x10
#DEFINE	YellowM 0x20

;  Macros
ClockStrobe MACRO               ;  Strobe the Clock Bit
  bsf    Clock
  bcf    Clock
 ENDM

EStrobe MACRO                   ;  Strobe the "E" Bit (Data & Q5)
  bsf    Data
  bcf    Data
 ENDM

 __CONFIG   _CP_OFF & _WDT_OFF & _MCLRE_OFF & _IntRC_OSC

  org    0

  goto Start

; ***************  Code Tables  ***********

GotoTable
  movf  cgoto_num, w
  addwf PCL

  goto  cgoto0
  goto  cgoto1
  goto  cgoto2
  goto  cgoto3
  goto  cgoto4
  goto  cgoto5
  goto  cgoto6
  goto  cgoto7
  goto  cgoto8
  goto  cgoto9
  goto  cgoto10
;  goto  cgoto11		;  not used (yet)
;  goto  cgoto12
;  goto  cgoto13

UserCharsDefTable		;  User chars to animation
  addwf  PCL                    

  retlw  0x00E                  ;  Character 0 (stop)
  retlw  0x00E
  retlw  0x004
  retlw  0x004
  retlw  0x004
  retlw  0x004
  retlw  0x00C
  retlw  0x000

  retlw  0x00E                  ;  Character 1
  retlw  0x00E
  retlw  0x004
  retlw  0x004
  retlw  0x00C
  retlw  0x004
  retlw  0x00C
  retlw  0x000

  retlw  0x00E                  ;  Character 2
  retlw  0x00E
  retlw  0x004
  retlw  0x004
  retlw  0x00A
  retlw  0x01A
  retlw  0x006
  retlw  0x000

  retlw  0x00E                  ;  Character 3
  retlw  0x00E
  retlw  0x004
  retlw  0x004
  retlw  0x006
  retlw  0x009
  retlw  0x019
  retlw  0x000

  retlw  0x00E                  ;  Character 4
  retlw  0x00E
  retlw  0x004
  retlw  0x004
  retlw  0x007
  retlw  0x009
  retlw  0x018
  retlw  0x000

  retlw  0x00E                  ;  Character 5
  retlw  0x00E
  retlw  0x004
  retlw  0x004
  retlw  0x00C
  retlw  0x006
  retlw  0x00C
  retlw	 0x000

Messages
  addwf	 PCL			;  PCL = $ + w

;        "0123456789012345"	;  line lenght = 16 (LM016)

Mess_Walk			;  Output table to line 1
   dt	1," WALKING ZUISTI", 0	;  dt = retlw char, ...

Mess_Yflash			;  Output table to line 2
   dt	 " YELLOW FLASHING", 0

Mess_Red			;  Output table to RED
   dt	 " STOP and WAIT !", 0

Mess_RedYell			;  Output table to RED-Yellow
   dt	 " .. GET READY ..", 0

Mess_Yell			;  Output table to Yellow
   dt	 "  ..  HURRY  .. ", 0

Mess_Green			;  Output table to Green
   dt	 " * FREE TO GO * ", 0

Mess_Anim			;  Output table to Animation
   dt	 1,2,3,4,5, 0

Mess_Shift			;  Output table to Shifting  (lenght=22)   
   dt	 32,32,32,3,32,32,32,3,32,32,32,3,32,32,32,3,32,32,32,3,32,32, 0
 

; ***********  callable subroutines  (used hw stack deep is max two) ******

SW2_AnimMode			; C reset if sw2 was grounded else set (1)

  movlw 0x2			; GPIO_1 is input temporaly
  TRIS	GPIO
  bcf STATUS, C
  btfsc GPIO, 1
   bsf STATUS, C
  clrw				; all (except GP3) is output again
  TRIS	GPIO
  retlw 0

SendCHAR                        ;  Send the Character to the LCD
  movwf  TempChar		;  Save

  swapf  TempChar, w		;  First send the high nybble
  call   NybbleOutC1

  movf   TempChar, w		;  Send the low nybble
NybbleOutC1
  bsf    STATUS, C		;  RS = 1
  goto   NybbleOut

GoHome
  movlw  0x080			;  Go to the first pos of the 1st line

SendINS                         ;  Send an instruction to the LCD
  movwf  TempChar		
  swapf  TempChar, w                
  call   NybbleOutC0
  movf   TempChar, w

NybbleOutC0
  bcf    STATUS, C		;  RS = 0

NybbleOut			;  Send a Nybble to the LCD

  movwf	 TempNybble		;  Save the nybble
  swapf  TempNybble		;  Setup to Output to the High Part of the Byte

  movlw  6                      ;  TO clear the shift Register, Delay1 is a cnt
  movwf  Delay1			
NO2Loop1
  ClockStrobe
  decfsz Delay1
   goto  NO2Loop1

  movlw  5                      ;  shift out the "RS" bit and the nybble 
  movwf  Delay1

  bsf    Data                   ;  Put out the LCD Gate Bit (E)
  ClockStrobe
NO2Loop2
  bcf    Data                   ;  Clear the Data pin
  btfsc  STATUS, C              ;  If the Bit (in C) to be output is a "1", set
   bsf   Data
  ClockStrobe
  rlf    TempNybble		;  Shift the next bit into the Carry (first D7)
  decfsz Delay1
   goto  NO2Loop2

  EStrobe                       ;  Strobe out the LCD Data

  retlw 0

Dlay160                         ;  Delay 160 usecs
  movlw ( 160 / 4 )     
  movwf Delay1	
  decfsz Delay1
   goto  $-1
  retlw 0

LongDlay
  clrw				;  Delay 256 * 5.9/8 = 192 mS
  goto DlayW

Dlay5                           ;  Delay cca 5.9 msecs
  movlw 8  
DlayW
  movwf  Delay1
  movwf  Delay2
  decfsz Delay2
   goto  $-1
  decfsz Delay1
   goto  $-3
  retlw 0


 if ( $ > 0x100 ) ; Address is max. 0x100 here
  ERROR "Tables and subroutines have exceeded page 0 (the first 256 loc.)"
 endif

; ************* Emulated call Subroutines  (control w. cgoto_num) ***********

; return with computed goto beacuse they are using the whole stack deep (2) yet 

MessageOut2
  movlw	 0x80 + 0x40		;  Go to the first pos of the 2nd line

MessageOutPosW
  call	 SendINS

MessageOut
  movf	 Offset, w
  call	 Messages		; get the next char
  iorlw	 0			; is it a zero end_sign?
  btfsc	 STATUS, Z
   goto	 GotoTable              ; if yes, return (comp. goto)
  call	 SendCHAR
  incf	 Offset
  goto MessageOut

AnimOut
  movlw  Mess_Anim - Messages - 1 
  movwf  Offset
Anim_loop
  call   GoHome
  movf	 Offset, w
  call	 Messages
  iorlw	 0
  btfsc	 STATUS, Z
   goto	 GotoTable
  call	 SendCHAR
  incf	 Offset
  movf   AnimSpeed, w
  call   DlayW
  goto   Anim_loop 

;*******************************************

Start

;C508 HARDWARE INIC

  clrf	GPIO			; clear all I/O pins
  clrw				; all except GP3 is output
  TRIS	GPIO 			; I/O directions set
  OPTION			; GP2 is an output (not TOCKI)

  movlw  32
  call   DlayW                  ; Wait 20 msecs before LCD inic

; LCD INIC

  movlw  0x03                   ;  Reset Command
  call   NybbleOutC0            ;  Send the Nybble with Clear Carry (instr)
  call   Dlay5                  ;  Wait 5 msecs before Sending again

  EStrobe
  call   Dlay160                ;  Wait 160 usecs before sending again 
;  call   Dlay5

  EStrobe
  call   Dlay160                ;  Wait 160 usecs before sending the next instr.
;  call   Dlay5

  movlw  0x02                   ;  Set 4 Bit Mode
  call   NybbleOutC0
  call   Dlay160
;  call   Dlay5

  movlw  0x028                  ;  It is a 2 Line Display
  call   SendINS

  movlw  0x008                  ;  Turn off the Display
  call   SendINS

  movlw  0x001                  ;  Clear the Display RAM
  call   SendINS
  call   Dlay5                  ;  can take up to 4.1 msecs

  movlw  0x006                  ;  Enable Cursor Move Direction
  call   SendINS

  movlw  0x00C                  ;  Turn the LCD Back On
  call   SendINS

  movlw  0x040                  ;  Put the Cursor to the Start of CGRAM
  call   SendINS

  movlw  48			;  Send 6 Characters of 8 Bytes
  movwf  AnimCount		;  Output the UserChar bytes

  clrf   Offset
UserCharLoop
  movf   Offset, w                 ;  Get the next offset to Output
  incf   Offset
  call   UserCharsDefTable
  call   SendCHAR               ;  Output the char
  decfsz AnimCount		;  Do 48x
   goto  UserCharLoop

DispRestart
  movlw 0x2			;  clear shifts if any
  call SendINS
  call Dlay5
  call GoHome
  movlw  Mess_Walk - Messages - 1  
  movwf  Offset
  movlw  0			; prepare for computed "return" 
  movwf  cgoto_num
  goto   MessageOut
cgoto0				; MessageOut returns here

MainLoop

  clrf   GPIO
  btfss  SW1_ContrMode
   goto  Red
  movlw  1
  movwf  cgoto_num
  movlw  Mess_Yflash - Messages - 1 
  movwf  Offset
  goto   MessageOut2
cgoto1

  movlw 180
  movwf AnimSpeed

Flashing

  movf   GPIO, w
  xorlw  YellowM
  movwf  GPIO			; toggle Yellow LED

  call  SW2_AnimMode
  btfss STATUS, C		; if sw2 was grounded (C=0), go to shifting
   goto  ShiftRoutine

  movlw 2
  movwf cgoto_num
  goto AnimOut
cgoto2
  btfsc  SW1_ContrMode		; if sw1 was grounded (0), go to normal mode
   goto  Flashing               ;   else again
   
Red
  clrf   GPIO			;  all leds off
  bsf   RedL			;  red led on	
  call  GoHome
  movlw 0			; "standing man" user char
  call SendCHAR
  movlw  7
  movwf  cgoto_num
  movlw  Mess_Red - Messages - 1 
  movwf  Offset
  goto   MessageOut2
cgoto7
  movlw  10
  movwf  Count1			; red loop timing
Rloop
  btfsc  SW1_ContrMode
   goto  MainLoop
  call	 LongDlay
  decfsz Count1
   goto Rloop

RedYellow
  bsf   YellowL	
  movlw  8
  movwf  cgoto_num
  movlw  Mess_RedYell - Messages - 1 
  movwf  Offset
  goto   MessageOut2
cgoto8
  movlw  9
  movwf  Count1			; RY loop timing
RYloop
  btfsc  SW1_ContrMode
   goto  MainLoop
  call	 LongDlay
  decfsz Count1
   goto  RYloop

Green
  clrf   GPIO
  bsf   GreenL	
  movlw  3
  movwf  cgoto_num
  movlw  Mess_Green - Messages - 1 
  movwf  Offset
  goto   MessageOut2
cgoto3

  movlw  4
  movwf  Count1			; green loop timing
  movlw  4
  movwf cgoto_num 
  movlw 150
  movwf AnimSpeed
Gloop
  btfsc  SW1_ContrMode
   goto  MainLoop
  goto AnimOut
cgoto4
  decfsz Count1
   goto  Gloop

Yellow
  clrf   GPIO
  bsf    YellowL	
  movlw  5
  movwf  cgoto_num
  movlw  Mess_Yell - Messages - 1 
  movwf  Offset
  goto   MessageOut2
cgoto5

  movlw  10
  movwf  Count1			; Yell loop timing
  movlw  6
  movwf  cgoto_num 
  movlw  60
  movwf  AnimSpeed
Yloop
  btfsc  SW1_ContrMode
   goto  MainLoop
  goto   AnimOut
cgoto6
  decfsz Count1
   goto  Yloop

  goto MainLoop

ShiftRoutine

  movlw 200
  movwf AnimSpeed

  call   GoHome
  movlw  9
  movwf  cgoto_num
  movlw  Mess_Shift - Messages - 1 
  movwf  Offset
  goto   MessageOut
cgoto9

ShiftLoop0
  movf   GPIO, w
  xorlw  YellowM
  movwf  GPIO			;  toggle Yellow LED

  movlw 0x80 + 0x41		;  line 2, pos at begin: first + 1 (bef. shift)
  movwf Count1			;  store as cursor pos

  movlw 4			;  4 shift step
  movwf Count

ShiftLoop1

  btfss  SW1_ContrMode
    goto DispRestart
  call  SW2_AnimMode
  btfsc STATUS, C		; if sw2 was grounded (C=0), skip  to shifting
    goto DispRestart


  movlw  10
  movwf  cgoto_num
  movlw  Mess_Shift - Messages - 1 + 2
  addwf  Count, w		;  modify stringtable pos to back shifting
  movwf  Offset
  movf	 Count1, w		;  w = new cursor pos
  goto   MessageOutPosW
cgoto10

  movlw 0x18 + 4	;chars shift left (display as a window shifts right)
  call SendINS

  movf AnimSpeed, w
  call DlayW

  incf Count1			;  incr cursor pos
  decfsz Count
    goto ShiftLoop1		;  next shift step

  movlw 0x2			;  clear shifts
  call SendINS
  movlw 4			;  long instr., min 2 mSec needed!
  call DlayW

  goto ShiftLoop0

  end
